"""
Infinite Minesweeper implemented with Python 3.12 and the textual library.

The game grid is composed of subgrids (8×8 each). The first (initial) subgrid
is generated at (0, 0) and new subgrids are generated on demand as you uncover cells.
Left-clicking a cell will reveal it (unless it is marked) and if it’s a mine the game ends.
Right-clicking or Shift + Left-clicking toggles a mine mark on that cell.
Only in the initial subgrid may any cell be clicked; in other subgrids only cells bordering an already uncovered cell are clickable.
"""

from __future__ import annotations

import os
import socketserver
from typing import Any

from rich.console import ConsoleRenderable, RichCast
from textual import on, work
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical
from textual.messages import ExitApp
from textual.visual import SupportsVisual, Visual
from textual.widgets import Footer, Header, Static

from par_infini_sweeper import __application_title__
from par_infini_sweeper.data_structures import GameState
from par_infini_sweeper.dialogs.difficulty_dialog import DifficultyDialog
from par_infini_sweeper.dialogs.help_dialog import HelpDialog
from par_infini_sweeper.dialogs.highscore_dialog import HighscoreDialog
from par_infini_sweeper.dialogs.login_dialog import AuthDialog
from par_infini_sweeper.dialogs.theme_dialog import ThemeDialog
from par_infini_sweeper.dialogs.url_dialog import UrlDialog
from par_infini_sweeper.enums import GameDifficulty
from par_infini_sweeper.main_grid import MainGrid
from par_infini_sweeper.messages import ShowURL, WebServerStarted, WebServerStopped


class PimApp(App):
    """
    Textual App for Infinite Minesweeper.
    Bindings:
      - n: New Game (prompts for difficulty)
      - h: Highscores
      - t: Change Theme
      - q: Quit
      - f1: Help
    """

    TITLE = __application_title__
    ENABLE_COMMAND_PALETTE = False
    ALLOW_SELECT = False
    CSS_PATH = "pim.tcss"
    BINDINGS = [
        Binding(key="n", action="new_game", description="New Game"),
        Binding(key="h", action="highscores", description="Highscores"),
        Binding(key="t", action="change_theme", description="Change Theme"),
        Binding(key="a", action="auth", description="Authentication"),
        Binding(key="q", action="quit", description="Quit"),
        Binding(key="f1", action="help", description="Help", show=True),
    ]

    def __init__(self, user_name: str | None = None, nickname: str | None = None, **kwargs: Any) -> None:
        if not user_name:
            user_name = os.environ.get("USER", "user")
        from par_infini_sweeper import db

        with db.get_db_connection() as conn:
            db.init_db(conn, user_name)

        super().__init__(**kwargs)
        self.info = Static("Info", id="info")
        self.debug_panel = Static("Debug", id="debug")
        self.game_state = GameState.load(None, user_name, nickname)
        self.sweeper_widget = MainGrid(self.game_state, self.info, self.debug_panel)
        self._web_server: socketserver.TCPServer | None = None

    def compose(self) -> ComposeResult:
        yield Header(show_clock=True)
        yield Footer()
        with Vertical():
            yield self.info
            with Horizontal():
                yield self.sweeper_widget
                yield self.debug_panel

    def on_mount(self) -> None:
        self.theme = self.game_state.theme
        self.sweeper_widget.focus()

    @work
    async def action_change_theme(self) -> None:
        """An action to change the theme."""
        self.game_state.theme = await self.push_screen_wait(ThemeDialog())
        self.game_state.save()

    def action_help(self) -> None:
        """Show help screen"""
        self.app.push_screen(HelpDialog())

    def action_highscores(self) -> None:
        """Display the highscores for each game mode."""
        self.app.push_screen(HighscoreDialog(self.game_state))

    def action_auth(self) -> None:
        """Display the auth dialog."""
        self.app.push_screen(AuthDialog(self.game_state))

    def set_debug(self, text: ConsoleRenderable | RichCast | str | SupportsVisual | Visual) -> None:
        self.debug_panel.update(text)

    @work
    async def action_new_game(self) -> None:
        """
        Start a new game. Prompts the user for a difficulty (easy, medium, or hard),
        resets the game state, and saves it.
        """
        difficulty: GameDifficulty | None = await self.push_screen_wait(DifficultyDialog())
        if difficulty is None:
            return
        self.game_state.difficulty = difficulty
        self.game_state.new_game()
        self.sweeper_widget.action_center()

    @on(ShowURL)
    def show_url(self, event: ShowURL) -> None:
        self.push_screen(
            UrlDialog(
                "OAUTH Login",
                "If your browser does not automatically open, click Copy and paste the url into a browser to start login.",
                event.url,
            )
        )

    @on(WebServerStarted)
    def webserver_started(self, event: WebServerStarted):
        self._web_server = event.server
        # stop the server if its still running after 5 minutes
        self.set_timer(300, self.stop_webserver)

    @on(WebServerStopped)
    def webserver_stopped(self):
        self._web_server = None

    def stop_webserver(self) -> None:
        if self._web_server:
            try:
                self._web_server.shutdown()
            except Exception as _:
                pass
            finally:
                self._web_server = None

    @on(ExitApp)
    def do_exit_app(self) -> None:
        self.stop_webserver()



"""Provides a dialog for getting difficulty level from the user."""

from __future__ import annotations

from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Center, Vertical
from textual.screen import ModalScreen
from textual.widgets import Button, Static

from par_infini_sweeper.enums import GameDifficulty


class DifficultyDialog(ModalScreen[GameDifficulty | None]):
    """Provides a dialog for getting difficulty level from the user."""

    DEFAULT_CSS = """
	DifficultyDialog {
		align: center middle;
	}

	DifficultyDialog > Vertical {
		background: $panel;
		height: auto;
		width: auto;
		border: thick $secondary;
		padding: 1;

        Center {
            width: 50;
        }
	}

	DifficultyDialog > Vertical > * {
		width: auto;
		height: auto;
	}

	DifficultyDialog Static {
		width: auto;
	}

	DifficultyDialog .spaced {
		padding: 1;
	}

	DifficultyDialog #question {
		min-width: 100%;
		border-top: solid $secondary;
		border-bottom: solid $secondary;
	}

	DifficultyDialog Button {
		margin-right: 1;
	}

	DifficultyDialog #buttons {
		width: 100%;
		align-horizontal: right;
		padding-right: 1;
	}
	"""

    BINDINGS = [
        Binding("left,up", "app.focus_previous", "", show=False),
        Binding("right,down", "app.focus_next", "", show=False),
        Binding("escape", "dismiss(None)", "", show=False),
    ]

    def __init__(
        self,
    ) -> None:
        """Initialise the yes/no dialog."""
        super().__init__()

    def compose(self) -> ComposeResult:
        """Compose the content of the dialog."""
        with Vertical():
            yield Static(
                "If you want to post your score to the internet,\nmake sure you do so before starting a new game!\n",
            )

            with Center():
                yield Static("Difficulty", classes="spaced")
                yield Button("Easy", id="easy")
                yield Button("Medium", id="medium")
                yield Button("Hard", id="hard")

    def on_mount(self) -> None:
        """Configure the dialog once the DOM is ready."""
        self.query(Button).first().focus()

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """Handle a button being pressed on the dialog."""
        self.dismiss(GameDifficulty(event.button.id))



from __future__ import annotations

from datetime import datetime
from typing import Literal

from pydantic import BaseModel, Field


class PimResult(BaseModel):
    status: Literal["success", "error"]
    message: str | None = None


class PostScoreResult(PimResult):
    pass


class PostScoreRequest(BaseModel):
    mode: Literal["infinite"]
    difficulty: Literal["easy", "medium", "hard"]
    score: int = Field(..., gt=0)
    duration: int = Field(..., gt=0)


class ScoreData(PostScoreRequest):
    nickname: str = Field(..., min_length=2, max_length=30)
    created_ts: datetime


class ScoreDataResponse(PostScoreResult):
    scores: list[ScoreData]


class ChangeNicknameRequest(BaseModel):
    nickname: str = Field(..., min_length=2, max_length=30)


class ChangeNicknameResponse(PimResult):
    pass



"""Par Infinite Minesweeper."""

from __future__ import annotations

import os

from dotenv import load_dotenv

load_dotenv()

__author__ = "Paul Robello"
__credits__ = ["Paul Robello"]
__maintainer__ = "Paul Robello"
__email__ = "probello@gmail.com"
__version__ = "0.3.7"
__application_title__ = "Par Infinite Minesweeper"
__application_binary__ = "pim"
__licence__ = "MIT"

from par_infini_sweeper.pim_app import PimApp  # noqa: E402

os.environ["USER_AGENT"] = f"{__application_title__} {__version__}"


__all__: list[str] = [
    "__author__",
    "__credits__",
    "__maintainer__",
    "__email__",
    "__version__",
    "__application_binary__",
    "__licence__",
    "__application_title__",
    "PimApp",
]



"""Provides a modal dialog for selecting a theme."""

from __future__ import annotations

from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical
from textual.screen import ModalScreen
from textual.widgets import Button, Input, RadioButton, RadioSet


class ThemeDialog(ModalScreen[str]):
    """Provides a modal dialog for selecting a theme."""

    # pylint: disable=duplicate-code
    DEFAULT_CSS = """
    ThemeDialog {
        align: center middle;
        background: transparent;
    }

    ThemeDialog > Vertical {
        background: $panel;
        height: auto;
        width: auto;
        border: thick $primary;
        border-title-color: $primary;
    }

    ThemeDialog > Vertical > * {
        width: auto;
        height: auto;
    }

    ThemeDialog RadioSet {
        width: 40;
        margin: 1;
    }

    ThemeDialog Label {
        margin-left: 2;
    }

    ThemeDialog Static {
        margin-left: 2;
    }

    ThemeDialog Button {
        margin-right: 1;
    }

    ThemeDialog #buttons {
        width: 100%;
        align-horizontal: right;
        padding-right: 1;
    }
    """

    BINDINGS = [
        Binding("escape", "cancel_input", "", show=False),
    ]

    def __init__(
        self,
    ) -> None:
        """Initialise the input dialog."""
        super().__init__()
        self.title = "Select Theme"
        self._initial_theme = self.app.theme
        self._themes = list(self.app.available_themes.keys())

    def compose(self) -> ComposeResult:
        """Compose the child widgets."""
        with Vertical() as v1:
            v1.border_title = self.title

            with RadioSet(id="theme_list"):
                yield from [RadioButton(theme, value=theme == self._initial_theme) for theme in self._themes]
            with Horizontal(id="buttons"):
                yield Button("OK", id="ok", variant="primary")
                yield Button("Cancel", id="cancel")

    def on_mount(self) -> None:
        """Set up the dialog once the DOM is ready."""
        self.query_one(RadioSet).focus()

    def on_radio_set_changed(self, event: RadioSet.Changed) -> None:
        """Set the theme."""
        if event.pressed.label.plain in self._themes:
            self.app.theme = event.pressed.label.plain

    def action_cancel_input(self) -> None:
        """Cancel the input operation."""
        self.cancel_input()

    @on(Button.Pressed, "#cancel")
    def cancel_input(self) -> None:
        """Cancel the input operation."""
        self.app.theme = self._initial_theme
        self.app.pop_screen()

    @on(Input.Submitted)
    @on(Button.Pressed, "#ok")
    def accept_input(self) -> None:
        """Accept and return the input."""
        self.dismiss(self.app.theme)



"""Provides an error dialog."""

from __future__ import annotations

from textual.widgets.button import ButtonVariant

from .text_dialog import TextDialog


class ErrorDialog(TextDialog):
    """Modal dialog for showing errors."""

    DEFAULT_CSS = """
    ErrorDialog > Vertical {
        background: $error 15%;
        border: thick $error 50%;
    }

    ErrorDialog #message {
        border-top: solid $panel;
        border-bottom: solid $panel;
    }
    """

    @property
    def button_style(self) -> ButtonVariant:
        """The style for the dialog's button."""
        return "error"



"""Provides a base modal dialog for showing text to the user."""

from __future__ import annotations

from rich.console import ConsoleRenderable, RichCast
from rich.text import TextType
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Center, Vertical
from textual.screen import ModalScreen
from textual.visual import SupportsVisual
from textual.widgets import Button, Static
from textual.widgets.button import ButtonVariant


class TextDialog(ModalScreen[None]):
    """Base modal dialog for showing information."""

    DEFAULT_CSS = """
	TextDialog {
		align: center middle;
	}

	TextDialog Center {
		width: 100%;
	}

	TextDialog > Vertical {
		background: $boost;
		min-width: 30%;
		width: auto;
		height: auto;
		border: round $primary;
	}

	TextDialog Static {
		width: auto;
	}

	TextDialog .spaced {
		padding: 1 4;
	}

	TextDialog #message {
		min-width: 100%;
	}
	"""

    BINDINGS = [
        Binding("escape", "dismiss(None)", "", show=False),
    ]

    def __init__(self, title: TextType, message: ConsoleRenderable | RichCast | str | SupportsVisual) -> None:  # noqa: F821
        """Initialise the dialog."""
        super().__init__()
        self._title = title
        self._message = message

    @property
    def button_style(self) -> ButtonVariant:
        """The style for the dialog's button."""
        return "primary"

    def compose(self) -> ComposeResult:
        """Compose the content of the modal dialog."""
        with Vertical():
            with Center():
                yield Static(self._title, classes="spaced")
            yield Static(self._message, id="message", classes="spaced")
            with Center(classes="spaced"):
                yield Button("OK", variant=self.button_style)

    def on_mount(self) -> None:
        """Configure the dialog once the DOM has loaded."""
        self.query_one(Button).focus()

    def on_button_pressed(self) -> None:
        """Handle the OK button being pressed."""
        self.dismiss(None)



"""Provides a dialog for getting a yes/no response from the user."""

from __future__ import annotations

from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Center, Horizontal, Vertical
from textual.screen import ModalScreen
from textual.widgets import Button, Static


class YesNoDialog(ModalScreen[bool]):
    """A dialog for asking a user a yes/no question."""

    DEFAULT_CSS = """
	YesNoDialog {
		align: center middle;
	}

	YesNoDialog > Vertical {
		background: $panel;
		height: auto;
		width: auto;
		min-width: 38;
		border: thick $secondary;
	}

	YesNoDialog > Vertical > * {
		width: auto;
		height: auto;
	}

	YesNoDialog Static {
		width: auto;
	}

	YesNoDialog .spaced {
		padding: 1;
	}

	YesNoDialog #question {
		min-width: 100%;
		border-top: solid $secondary;
		border-bottom: solid $secondary;
	}

	YesNoDialog Button {
		margin-right: 1;
	}

	YesNoDialog #buttons {
		width: 100%;
		align-horizontal: right;
		padding-right: 1;
	}
	"""

    BINDINGS = [
        Binding("left,up", "app.focus_previous", "", show=False),
        Binding("right,down", "app.focus_next", "", show=False),
        Binding("escape", "dismiss(False)", "", show=False),
    ]

    def __init__(  # pylint:disable=too-many-arguments, too-many-positional-arguments
        self,
        title: str,
        question: str,
        yes_label: str = "Yes",
        no_label: str = "No",
        yes_first: bool = True,
    ) -> None:
        """Initialise the yes/no dialog.

        Args:
                title: The title for the dialog.
                question: The question to ask.
                yes_label: The optional label for the yes button.
                no_label: The optional label for the no button.
                yes_first: Should the yes button come first?
        """
        super().__init__()
        self._title = title
        self._question = question
        self._yes = yes_label
        self._no = no_label
        self._yes_first = yes_first

    def compose(self) -> ComposeResult:
        """Compose the content of the dialog."""
        with Vertical():
            with Center():
                yield Static(self._title, classes="spaced")
            yield Static(self._question, id="question", classes="spaced")
            with Horizontal(id="buttons"):
                aye = Button(self._yes, id="yes")
                naw = Button(self._no, id="no")
                if self._yes_first:
                    aye.variant = "primary"
                    yield aye
                    yield naw
                else:
                    naw.variant = "primary"
                    yield naw
                    yield aye

    def on_mount(self) -> None:
        """Configure the dialog once the DOM is ready."""
        self.query(Button).first().focus()

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """Handle a button being pressed on the dialog."""
        self.dismiss(event.button.id == "yes")



"""Provides a base modal dialog for showing text to the user."""

from __future__ import annotations

from rich.console import ConsoleRenderable, RichCast
from rich.text import TextType
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Center, Vertical
from textual.screen import ModalScreen
from textual.visual import SupportsVisual
from textual.widgets import Button, Static


class UrlDialog(ModalScreen[None]):
    """Base modal dialog for showing information."""

    DEFAULT_CSS = """
	UrlDialog {
		align: center middle;
	}

	UrlDialog Center {
		width: 100%;
	}

	UrlDialog > Vertical {
		background: $boost;
		min-width: 30%;
		width: auto;
		height: auto;
		border: round $primary;
	}

	UrlDialog Static {
		width: auto;
	}

	UrlDialog .spaced {
		padding: 1 4;
	}

	UrlDialog #message {
		min-width: 100%;
	}
	"""

    BINDINGS = [
        Binding("escape", "dismiss(None)", "", show=False),
    ]

    def __init__(self, title: TextType, message: ConsoleRenderable | RichCast | str | SupportsVisual, url: str) -> None:  # noqa: F821
        """Initialise the dialog."""
        super().__init__()
        self._title = title
        self._message = message
        self._url = url

    def compose(self) -> ComposeResult:
        """Compose the content of the modal dialog."""
        with Vertical():
            with Center():
                yield Static(self._title, classes="spaced")
            yield Static(self._message, id="message", classes="spaced")
            with Center(classes="spaced"):
                yield Button("Copy", variant="success")

    def on_mount(self) -> None:
        """Configure the dialog once the DOM has loaded."""
        self.query_one(Button).focus()

    def on_button_pressed(self) -> None:
        """Handle the OK button being pressed."""
        self.app.copy_to_clipboard(self._url)
        self.notify("URL copied to clipboard")
        self.dismiss(None)



from __future__ import annotations

from enum import StrEnum


class GameMode(StrEnum):
    INFINITE = "infinite"


class GameDifficulty(StrEnum):
    EASY = "easy"
    MEDIUM = "medium"
    HARD = "hard"



"""Main application"""

from __future__ import annotations

import os
from typing import Annotated

import typer
from rich.console import Console
from textual_serve.server import Server

from par_infini_sweeper import __application_title__, __version__
from par_infini_sweeper.pim_app import PimApp

app = typer.Typer()
console = Console(stderr=True)


def version_callback(value: bool) -> None:
    """Print version and exit."""
    if value:
        print(f"{__application_title__}: {__version__}")
        raise typer.Exit()


@app.command()
def main(
    start_server: Annotated[
        bool, typer.Option("--server", "-s", help="Start webserver that allows app to be played in a browser")
    ] = False,
    user_name: Annotated[str, typer.Option("--user", "-u", help="User name to use")] = os.environ.get("USER", "user"),
    nickname: Annotated[str | None, typer.Option("--nick", "-n", help="Set user nickname")] = None,
    version: Annotated[  # pylint: disable=unused-argument
        bool | None,
        typer.Option("--version", "-v", callback=version_callback, is_eager=True),
    ] = None,
) -> None:
    """Main function."""

    if user_name and len(user_name) > 20:
        console.print("User name must be 20 characters or less")
        raise typer.Exit(1)
    if nickname and len(nickname) > 20:
        console.print("Nickname must be 20 characters or less")
        raise typer.Exit(1)

    if start_server:
        server_args: list[str] = ["pim"]
        if user_name:
            server_args.extend(["--user", user_name])
        if nickname:
            server_args.extend(["--nick", nickname])
        server = Server(" ".join(server_args))
        server.serve()
        return

    sweeper_app: PimApp = PimApp(user_name, nickname)
    sweeper_app.run()


if __name__ == "__main__":
    main_app = app()



from __future__ import annotations

import http.server
import os
import socketserver
import threading
import webbrowser
from typing import Any

from authlib.integrations.requests_client import OAuth2Session
from textual.app import App

from par_infini_sweeper import db
from par_infini_sweeper.messages import ShowURL, WebServerStarted, WebServerStopped

AUTH0_DOMAIN = os.environ.get("AUTH0_DOMAIN", "par-dev.us.auth0.com")
AUTH0_CLIENT_ID = os.environ.get("AUTH0_CLIENT_ID", "79LqAgzljQjin3dihidWdmLSD7o4CpL5")
API_AUDIENCE = "PIM"
REDIRECT_URI = "http://127.0.0.1:1999/oauth/callback"


def build_auth_client(user: dict[str, Any], app: App) -> OAuth2Session:
    """
    Returns a rest client that sends authorization bearer token with requests

    Args:
        user: (dict) User dictionary with token info
        app: (App) Textual app
    """
    token_data = {
        "access_token": user["access_token"] or None,
        "refresh_token": user["refresh_token"] or None,
        "expires_at": user["expires_at"] or None,
        "token_type": "Bearer",
    }

    client = OAuth2Session(
        client_id=AUTH0_CLIENT_ID,
        redirect_uri=REDIRECT_URI,
        scope="openid profile email offline_access",
        token=token_data,
    )
    if client.token:
        # if auth token does not exist or is expired and we have a refresh token, refresh it
        if client.token.is_expired() in [None, True] and client.token["refresh_token"]:
            try:
                refresh_token_url = f"https://{AUTH0_DOMAIN}/oauth/token"

                new_token = client.refresh_token(
                    url=refresh_token_url,  # Use the correct URL here
                    refresh_token=client.token["refresh_token"],
                )

                client.token = new_token

                user["access_token"] = client.token["access_token"]
                user["refresh_token"] = client.token["refresh_token"]
                user["expires_at"] = client.token["expires_at"]
                db.save_user(user)
                return client
            except Exception as _:
                # refresh failed, blank out auth info and start fresh
                # print(f"Error refreshing token: {e}")
                user["access_token"] = ""
                user["refresh_token"] = ""
                user["expires_at"] = 0
                db.save_user(user)
        elif client.token["access_token"]:
            # we have an auth token and its not expired
            return client

    return start_auth_server(client, user, app)


def start_auth_server(client: OAuth2Session, user: dict[str, Any], app: App) -> OAuth2Session:
    def login() -> Any:
        """
        Initiate Auth0 OAuth login by redirecting the user to the Auth0 login page.
        """
        authorization_url, state = client.create_authorization_url(
            f"https://{AUTH0_DOMAIN}/authorize", audience=API_AUDIENCE
        )
        if not webbrowser.open(authorization_url):
            app.post_message(ShowURL(authorization_url))

    class OAuthCallbackHandler(http.server.SimpleHTTPRequestHandler):
        def do_GET(self):
            """Handle the OAuth callback from Auth0.
            This function exchanges the authorization code for tokens and stores the tokens in the db.
            """
            # Allow any request path that starts with "/oauth/callback"
            if not self.path.startswith("/oauth/callback"):
                self.send_error(404, "Not Found")
            try:
                # Include query parameters in the authorization_response URL
                authorization_response = f"http://{self.headers['Host']}{self.path}"
                token = client.fetch_token(
                    f"https://{AUTH0_DOMAIN}/oauth/token",
                    authorization_response=authorization_response,
                )
                user["id_token"] = token.get("id_token")
                user["access_token"] = token.get("access_token")
                user["refresh_token"] = token.get("refresh_token")
                user["expires_at"] = token.get("expires_at")

                db.save_user(user)
            except Exception as e:
                self.send_error(500, str(e))
                return

            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.end_headers()
            self.wfile.write(b"Logged in. You may close this window.")

            threading.Thread(target=self.shutdown_server).start()

        def shutdown_server(self):
            self.server.shutdown()

    with socketserver.TCPServer(("127.0.0.1", 1999), OAuthCallbackHandler) as httpd:
        try:
            app.post_message(WebServerStarted(httpd))
            login()
            httpd.serve_forever()
        finally:
            app.post_message(WebServerStopped())

    return client



"""Provides a base modal dialog for showing text to the user."""

from __future__ import annotations

from textual import on, work
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Vertical, VerticalScroll
from textual.screen import ModalScreen
from textual.widgets import Button, DataTable, Footer, TabbedContent

from par_infini_sweeper import db
from par_infini_sweeper.data_structures import GameState
from par_infini_sweeper.dialogs.login_dialog import AuthDialog
from par_infini_sweeper.enums import GameMode
from par_infini_sweeper.utils import escape_brackets, format_duration


class HighscoreDialog(ModalScreen[None]):
    """Base modal dialog for showing information."""

    DEFAULT_CSS = """
	HighscoreDialog {
		align: center middle;
	}

	#main {
		background: $boost;
		width: 1fr;
		height: 1fr;
		border: round $primary;
	}
	#tc {
	}

	HighscoreDialog .spaced {
		padding: 1 4;
	}

	HighscoreDialog .scores {
		min-width: 100%;
	}
	#local_vs {
		width: 1fr;
		height: 1fr;
	}
	DataTable {
	    border: round $primary;
	}
	"""

    BINDINGS = [
        Binding("r", "refresh", "Refresh", show=True),
        Binding("l", "local", "Local", show=True),
        Binding("i", "internet", "Internet", show=True),
        Binding("p", "post_score", "Post Score", show=True),
        Binding("q,escape", "dismiss(None)", "Return", show=True),
    ]

    def __init__(self, game_state: GameState, show_internet: bool = False) -> None:  # noqa: F821
        """Initialise the dialog."""
        super().__init__()
        self.game_state = game_state
        self.tc = TabbedContent("Local", "Internet", id="tc")
        self.show_internet = show_internet

    @work
    async def load_local_data(self) -> None:
        """Load local data."""
        self.query_exactly_one("#local_vs", VerticalScroll).loading = True
        try:
            data = db.get_highscores()
            for mode, data in data.items():
                rows = [("Name", "Score", "Time", "Submitted")]
                for row in data:
                    rows.append(
                        (row["nickname"], str(row["score"]), format_duration(row["duration"]), row["created_ts"])
                    )
                table = self.query_exactly_one(f"#local_dt_{mode}", DataTable)
                table.clear(True)
                table.add_columns(*rows[0])
                table.add_rows(rows[1:])
        except Exception as _:
            self.notify("Error loading remote data", severity="error")
        finally:
            self.query_exactly_one("#local_vs", VerticalScroll).loading = False

    def compose_local(self) -> ComposeResult:
        with VerticalScroll(id="local_vs") as vs:
            vs.loading = True
            for mode in GameMode:
                dt = DataTable(
                    id=f"local_dt_{mode}",
                    zebra_stripes=True,
                    show_cursor=False,
                    cursor_type="row",
                    classes="spaced scores",
                )
                dt.border_title = mode.name.capitalize()
                yield dt

    @work
    async def load_remote_data(self) -> None:
        """Load internet data."""
        self.query_exactly_one("#internet_vs", VerticalScroll).loading = True
        try:
            data = db.get_internet_highscores()
            # print(data)
            for mode, data in data.items():
                rows = [("Name", "Score", "Time", "Submitted")]
                for row in data:
                    rows.append(
                        (
                            row.nickname,
                            str(row.score),
                            format_duration(row.duration),
                            row.created_ts.isoformat().split(".")[0],
                        )
                    )
                table = self.query_exactly_one(f"#dt_{mode}", DataTable)
                table.clear(True)
                table.add_columns(*rows[0])
                table.add_rows(rows[1:])
                self.notify("Loaded remote data")
        except Exception as e:
            # self.notify("Error loading remote data", severity="error")
            self.notify(escape_brackets(str(e)), severity="error")
        finally:
            self.query_exactly_one("#internet_vs", VerticalScroll).loading = False

    def compose_internet(self) -> ComposeResult:
        with VerticalScroll(id="internet_vs") as vs:
            yield Button("Post Score", id="post_score")
            vs.loading = True
            for mode in GameMode:
                dt = DataTable(
                    id=f"dt_{mode}",
                    zebra_stripes=True,
                    show_cursor=False,
                    cursor_type="row",
                    classes="spaced scores",
                )
                dt.border_title = mode.name.capitalize()
                yield dt

    def compose(self) -> ComposeResult:
        """Compose the content of the modal dialog."""
        yield Footer()
        with Vertical(id="main"):
            with self.tc:
                yield from self.compose_local()
                yield from self.compose_internet()

    def action_refresh(self) -> None:
        """Refresh the data in the dialog."""
        self.load_local_data()
        self.load_remote_data()

    def action_internet(self) -> None:
        """Change to internet tab"""
        self.tc.active = "tab-2"

    def action_local(self) -> None:
        """Change to local tab"""
        self.tc.active = "tab-1"

    def on_mount(self) -> None:
        """Configure the dialog once the DOM has loaded."""
        self.query_one(DataTable).focus()
        self.load_local_data()
        if self.show_internet:
            self.tc.active = "tab-2"
        # self.load_remote_data()

    @on(TabbedContent.TabActivated)
    def on_remote_tab_selected(self, event: TabbedContent.TabActivated) -> None:
        """Load the remote data when the tab is selected."""
        # self.notify(event.tab.id or "?")
        if event.tab.id == "--content-tab-tab-2":
            self.load_remote_data()

    @on(Button.Pressed, "#post_score")
    def action_post_score(self) -> None:
        self.app.push_screen(AuthDialog(self.game_state))



"""The main help dialog for the application."""

from __future__ import annotations

from pathlib import Path
from typing import Final

from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Center, Vertical
from textual.screen import ModalScreen
from textual.widgets import Button, Markdown

from par_infini_sweeper import __application_title__, __version__

HELP: Final[str] = f"""
# {__application_title__} v{__version__} Help

"""


class HelpDialog(ModalScreen[None]):
    """Modal dialog that shows the application's help."""

    DEFAULT_CSS = """
    HelpDialog {
        align: center middle;
        & > Vertical {
            border: thick $primary 50%;
            width: 80%;
            height: 80%;
            background: $boost;
            Markdown {
                height: 1fr;
            }
        }
    }
    """

    BINDINGS = [
        Binding("escape,f1", "dismiss(None)", "", show=False),
    ]

    def compose(self) -> ComposeResult:
        """Compose the help screen."""
        help_text = (Path(__file__).parent.parent / "help.md").read_text(encoding="utf-8")
        with Vertical():
            yield Markdown(HELP + help_text)
            with Center():
                yield Button("Close", variant="primary")

    def on_mount(self) -> None:
        """Configure the help screen once the DOM is ready."""
        # It seems that some things inside Markdown can still grab focus;
        # which might not be right. Let's ensure that can't happen here.
        self.query_one(Markdown).can_focus_children = False
        self.query_one(Button).focus()

    def on_button_pressed(self) -> None:
        """React to button press."""
        self.dismiss(None)

    def on_markdown_link_clicked(self, event: Markdown.LinkClicked) -> None:
        """A link was clicked in the help."""
        self.app.open_url(event.href)



"""Provides a modal dialog for getting a masked value from the user."""

from __future__ import annotations

from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical
from textual.screen import ModalScreen
from textual.widgets import Button, Input, Label


class PasswordDialog(ModalScreen[str]):
    """Provides a modal dialog for getting a masked value from the user."""

    DEFAULT_CSS = """
    InputDialog {
        align: center middle;
    }

    InputDialog > Vertical {
        background: $panel;
        height: auto;
        width: auto;
        border: thick $primary;
    }

    InputDialog > Vertical > * {
        width: auto;
        height: auto;
    }

    InputDialog Input {
        width: 40;
        margin: 1;
    }

    InputDialog Label {
        margin-left: 2;
    }

    InputDialog Button {
        margin-right: 1;
    }

    InputDialog #buttons {
        width: 100%;
        align-horizontal: right;
        padding-right: 1;
    }
    """

    BINDINGS = [
        Binding("escape", "app.pop_screen", "", show=False),
    ]
    """Bindings for the dialog."""

    def __init__(self, prompt: str, initial: str | None = None) -> None:
        """Initialise the dialog."""
        super().__init__()
        self._prompt = prompt
        self._initial = initial

    def compose(self) -> ComposeResult:
        """Compose the child widgets."""
        with Vertical():
            with Vertical(id="input"):
                yield Label(self._prompt)
                yield Input(self._initial or "", password=True)
            with Horizontal(id="buttons"):
                yield Button("OK", id="ok", variant="primary")
                yield Button("Cancel", id="cancel")

    def on_mount(self) -> None:
        """Set up the dialog once the DOM is ready."""
        self.query_one(Input).focus()

    @on(Button.Pressed, "#cancel")
    def cancel_input(self) -> None:
        """Cancel the input operation."""
        self.app.pop_screen()

    @on(Input.Submitted)
    @on(Button.Pressed, "#ok")
    def accept_input(self) -> None:
        """Accept and return the input."""
        if value := self.query_one(Input).value.strip():
            self.dismiss(value)



"""Provides login / out / nickname management dialog."""

from __future__ import annotations

from dataclasses import dataclass
from textwrap import dedent
from typing import Literal

from textual import on, work
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical
from textual.message import Message
from textual.screen import ModalScreen
from textual.widgets import Button, Footer, Label, Static

from par_infini_sweeper.data_structures import GameState
from par_infini_sweeper.dialogs.input_dialog import InputDialog
from par_infini_sweeper.utils import escape_brackets


class AuthDialog(ModalScreen[None]):
    """Provides login / out / nickname management dialog."""

    DEFAULT_CSS = """
	AuthDialog {
		align: center middle;
		.spaced {
		    padding: 1 4;
	    }
	}

	#main {
	    border-title-color: red;
		border: round $primary;
		background: $boost;
		width: 1fr;
		height: 1fr;
		padding: 1;
		Horizontal {
		    width: 1fr;
		    height: 3;
		    align: left middle;
		    Label {
		        margin: 1 1 0 0;
		    }
        }
	}

	#blurb {
		width: 1fr;
		margin: 0 0 1 0;
	}
	#net_nickname{
	    margin-right: 1;
	}
	Button {
	    margin-bottom: 1;
	}
	"""

    BINDINGS = [
        Binding("q,escape", "dismiss(None)", "Return", show=True),
    ]

    @dataclass
    class OauthDone(Message):
        status: Literal["success", "error"]
        msg: str

    def __init__(self, game_state: GameState) -> None:  # noqa: F821
        """Initialise the dialog."""
        super().__init__()

        self.game_state = game_state

    @work(thread=True, exclusive=True, group="login")
    async def do_login(self) -> None:
        """Do OAUTH2 flow."""
        try:
            if self.game_state.auth_client:
                self.post_message(AuthDialog.OauthDone("success", "Logged in"))
        except Exception as e:
            self.notify(str(e), severity="error")
            self.post_message(AuthDialog.OauthDone("error", str(e)))

    @on(OauthDone)
    async def oauth_done(self, event: OauthDone) -> None:
        """OAUTH done."""
        self.query_exactly_one("#main").loading = False
        if event.status == "success":
            self.notify(event.msg)
        else:
            self.notify(event.msg, severity="error")
        await self.recompose()

    def compose(self) -> ComposeResult:
        """Compose the content of the modal dialog."""
        logged_in = self.game_state.is_logged_in()

        yield Footer()
        with Vertical(id="main") as v:
            if self.game_state.game_over:
                v.border_title = "Game Over"
            with Horizontal():
                if logged_in:
                    yield Label("Net Nickname: ", id="net_nick_label")
                    yield Label(self.game_state.user["net_nickname"] or "Not set", id="net_nickname")
                    nickname_button = Button.warning(
                        "Change Net Nickname" if self.game_state.user["net_nickname"] else "Register Net Nickname",
                        id="change_nickname",
                    )
                    yield nickname_button
                else:
                    yield Static("\n[yellow]Login to set nickname")

            yield Static(
                dedent("""
                Clicking Login uses an OAUTH social provider such as Google, Facebook, Github.
                See the Readme or full help if you want to know more about how this works and our data policies."""),
                id="blurb",
            )
            login_button = Button.success("Login", id="login")
            logout_button = Button.error("Logout", id="logout")
            login_button.display = not logged_in
            logout_button.display = not login_button.display
            yield login_button
            yield logout_button
            if not self.game_state.score():
                yield Static("No score to upload", id="score_label")
            else:
                yield Static(
                    " - ".join(
                        [
                            f"Difficulty: [#00FF00]{self.game_state.difficulty}[/]",
                            f"Solved: {self.game_state.num_solved} / {self.game_state.num_subgrids}",
                            f"Score: [#00FF00]{self.game_state.score()}[/]",
                            f"Time: {self.game_state.time_played}\n",
                        ]
                    )
                )
                if not logged_in:
                    yield Static("\n[yellow]Login to upload score")
                else:
                    yield Button.success("Upload Score", id="upload")

    @on(Button.Pressed, "#login")
    def login_click(self) -> None:
        """Login button clicked"""
        self.query_exactly_one("#main").loading = True
        self.do_login()

    @on(Button.Pressed, "#logout")
    async def logout_click(self) -> None:
        """Logout button clicked"""
        self.game_state.logout()
        await self.recompose()

    @on(Button.Pressed, "#upload")
    async def upload_click(self) -> None:
        """Upload button clicked"""
        if not self.game_state.is_logged_in():
            self.notify("Not logged in", severity="error")
            return
        try:
            result = self.game_state.post_internet_score()
            if result.status != "success":
                raise Exception(result.message)
            self.notify("Score uploaded", severity="information")
        except Exception as e:
            self.notify(escape_brackets(str(e)), severity="error")

    @on(Button.Pressed, "#change_nickname")
    @work
    async def change_nickname_click(self) -> None:
        """Change net_nickname button clicked"""
        new_nickname = await self.app.push_screen_wait(
            InputDialog(
                "New Nickname",
                self.game_state.user["net_nickname"] or self.game_state.user["nickname"],
                "Change Nickname",
                "Name may only contain letters, numbers and . - _  \n",
            )
        )
        if not new_nickname or len(new_nickname) < 2 or len(new_nickname) > 30:
            self.notify("Nickname must be between 2 and 30 characters", severity="error")
            return
        try:
            result = self.game_state.change_internet_nickname(new_nickname)
            if result.status != "success":
                raise Exception(result.message)
            else:
                self.notify(escape_brackets(result.message or "Nickname changed"), severity="information")
        except Exception as e:
            self.notify(escape_brackets(str(e)), severity="error")
            return
        await self.recompose()

    def on_mount(self) -> None:
        """Configure the dialog once the DOM has loaded."""



from sqlite3 import Connection


def migrate_db_to_1_1(conn: Connection) -> None:
    """
    Migrate the SQLite database from version 1.0 to 1.1.

    Args:
        conn (Connection): SQLite connection object.
    """
    with conn:
        cursor = conn.cursor()

        # Add created_ts column to users table if it doesn't exist
        cursor.execute("PRAGMA table_info(games)")
        columns = [col[1] for col in cursor.fetchall()]

        if "duration" not in columns:
            cursor.execute("ALTER TABLE games ADD COLUMN duration INTEGER NOT NULL DEFAULT 0")
            cursor.execute("UPDATE games SET duration = play_duration")
            cursor.execute("ALTER TABLE games DROP COLUMN play_duration")

        if "mode" not in columns:
            cursor.execute("ALTER TABLE games ADD COLUMN mode TEXT NOT NULL DEFAULT 'infinite'")
            cursor.execute("UPDATE games SET mode = game_mode")
            cursor.execute("ALTER TABLE games DROP COLUMN game_mode")

        cursor.execute("PRAGMA table_info(users)")
        columns = [col[1] for col in cursor.fetchall()]
        if "net_nickname" not in columns:
            cursor.execute("ALTER TABLE users ADD COLUMN net_nickname TEXT DEFAULT ''")
            cursor.execute("UPDATE users SET net_nickname = ''")

        if "id_token" not in columns:
            cursor.execute("ALTER TABLE users ADD COLUMN id_token TEXT DEFAULT ''")
            cursor.execute("UPDATE users SET id_token = ''")
        if "access_token" not in columns:
            cursor.execute("ALTER TABLE users ADD COLUMN access_token TEXT DEFAULT ''")
            cursor.execute("UPDATE users SET access_token = ''")
        if "refresh_token" not in columns:
            cursor.execute("ALTER TABLE users ADD COLUMN refresh_token TEXT DEFAULT ''")
            cursor.execute("UPDATE users SET refresh_token = ''")
        if "expires_at" not in columns:
            cursor.execute("ALTER TABLE users ADD COLUMN expires_at INTEGER DEFAULT 0")
            cursor.execute("UPDATE users SET expires_at = 0")

        cursor.execute("UPDATE pim_db_info set version = ?", ("1.1",))


def migrate_legacy_db(conn: Connection) -> None:
    """
    Migrate the SQLite database to the current schema.

    Args:
        conn (Connection): SQLite connection object.
    """
    with conn:
        cursor = conn.cursor()

        # Add created_ts column to users table if it doesn't exist
        cursor.execute("PRAGMA table_info(users)")
        columns = [col[1] for col in cursor.fetchall()]
        if "created_ts" not in columns:
            cursor.execute("ALTER TABLE users ADD COLUMN created_ts TIMESTAMP")
            cursor.execute("UPDATE users SET created_ts = CURRENT_TIMESTAMP WHERE created_ts IS NULL")

        # Add game_mode column to games table if it doesn't exist
        cursor.execute("PRAGMA table_info(games)")
        columns = [col[1] for col in cursor.fetchall()]
        if "game_mode" not in columns:
            cursor.execute("ALTER TABLE games ADD COLUMN game_mode TEXT NOT NULL DEFAULT 'infinite'")

        # Add ON DELETE CASCADE to foreign keys if not present
        cursor.execute("PRAGMA foreign_key_list(user_prefs)")
        foreign_keys = cursor.fetchall()
        if not any(fk[5] == "CASCADE" for fk in foreign_keys):
            cursor.execute("""
                CREATE TABLE user_prefs_new (
                    id INTEGER PRIMARY KEY,
                    theme TEXT NOT NULL,
                    difficulty TEXT NOT NULL CHECK(difficulty IN ('easy','medium','hard')),
                    FOREIGN KEY(id) REFERENCES users(id) ON DELETE CASCADE
                )
            """)
            cursor.execute("INSERT INTO user_prefs_new SELECT * FROM user_prefs")
            cursor.execute("DROP TABLE user_prefs")
            cursor.execute("ALTER TABLE user_prefs_new RENAME TO user_prefs")

        cursor.execute("PRAGMA foreign_key_list(games)")
        foreign_keys = cursor.fetchall()
        if not any(fk[5] == "CASCADE" for fk in foreign_keys):
            cursor.execute("""
                CREATE TABLE games_new (
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    user_id INTEGER NOT NULL,
                    game_mode TEXT NOT NULL DEFAULT 'infinite',
                    game_over BOOLEAN NOT NULL DEFAULT 0,
                    duration INTEGER NOT NULL DEFAULT 0,
                    board_offset TEXT NOT NULL DEFAULT '0,0',
                    created_ts TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                    FOREIGN KEY(user_id) REFERENCES users(id) ON DELETE CASCADE
                )
            """)
            cursor.execute(
                "INSERT INTO games_new (id, user_id, game_mode, game_over, duration, board_offset) SELECT id, user_id, game_mode, game_over, duration, board_offset FROM games"
            )
            cursor.execute("DROP TABLE games")
            cursor.execute("ALTER TABLE games_new RENAME TO games")

        cursor.execute("PRAGMA foreign_key_list(highscores)")
        foreign_keys = cursor.fetchall()
        if not any(fk[5] == "CASCADE" for fk in foreign_keys):
            cursor.execute("""
                CREATE TABLE highscores_new (
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    user_id INTEGER NOT NULL,
                    game_id INTEGER NOT NULL,
                    score INTEGER NOT NULL,
                    created_ts TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                    FOREIGN KEY(user_id) REFERENCES users(id) ON DELETE CASCADE,
                    FOREIGN KEY(game_id) REFERENCES games(id) ON DELETE CASCADE
                )
            """)
            cursor.execute("INSERT INTO highscores_new SELECT * FROM highscores")
            cursor.execute("DROP TABLE highscores")
            cursor.execute("ALTER TABLE highscores_new RENAME TO highscores")

        cursor.execute("PRAGMA foreign_key_list(grids)")
        foreign_keys = cursor.fetchall()
        if not any(fk[5] == "CASCADE" for fk in foreign_keys):
            cursor.execute("""
                CREATE TABLE grids_new (
                    game_id INTEGER NOT NULL,
                    user_id INTEGER NOT NULL,
                    sub_grid_id TEXT NOT NULL,
                    grid_data TEXT NOT NULL,
                    FOREIGN KEY(user_id) REFERENCES users(id) ON DELETE CASCADE,
                    FOREIGN KEY(game_id) REFERENCES games(id) ON DELETE CASCADE,
                    PRIMARY KEY (game_id, user_id, sub_grid_id)
                )
            """)
            cursor.execute("INSERT INTO grids_new SELECT * FROM grids")
            cursor.execute("DROP TABLE grids")
            cursor.execute("ALTER TABLE grids_new RENAME TO grids")



"""Provides a modal dialog for getting a value from the user."""

from __future__ import annotations

from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical
from textual.screen import ModalScreen
from textual.widgets import Button, Input, Label, Static


class InputDialog(ModalScreen[str]):
    """A modal dialog for getting a single input from the user."""

    # pylint: disable=duplicate-code
    DEFAULT_CSS = """
    InputDialog {
        align: center middle;
    }

    InputDialog > Vertical {
        background: $panel;
        height: auto;
        width: auto;
        border: thick $primary;
    }

    InputDialog > Vertical > * {
        width: auto;
        height: auto;
    }

    InputDialog Input {
        width: 40;
        margin: 1;
    }

    InputDialog Label {
        margin-left: 2;
    }

    InputDialog Static {
        margin-left: 2;
    }

    InputDialog Button {
        margin-right: 1;
    }

    InputDialog #buttons {
        width: 100%;
        align-horizontal: right;
        padding-right: 1;
    }
    """

    BINDINGS = [
        Binding("escape", "screen.dismiss('')", "", show=False),
    ]

    def __init__(self, prompt: str, initial: str | None = None, title: str = "", msg: str = "") -> None:
        """Initialise the input dialog.

        Args:
                prompt: The prompt for the input.
                initial: The initial value for the input.
                title: The title for the dialog.
                msg: The message to show.
        """
        super().__init__()
        self.title = title
        self.msg = msg
        self._prompt = prompt
        self._initial = initial

    def compose(self) -> ComposeResult:
        """Compose the child widgets."""
        v1 = Vertical()
        if self.title:
            v1.border_title = self.title
        with v1:
            if self.msg:
                yield Static(self.msg)
            with Vertical(id="input"):
                yield Label(self._prompt)
                yield Input(self._initial or "")
            with Horizontal(id="buttons"):
                yield Button("OK", id="ok", variant="primary")
                yield Button("Cancel", id="cancel")

    def on_mount(self) -> None:
        """Set up the dialog once the DOM is ready."""
        self.query_one(Input).focus()

    @on(Button.Pressed, "#cancel")
    def cancel_input(self) -> None:
        """Cancel the input operation."""
        self.app.pop_screen()

    @on(Input.Submitted)
    @on(Button.Pressed, "#ok")
    def accept_input(self) -> None:
        """Accept and return the input."""
        if value := self.query_one(Input).value.strip():
            self.dismiss(value)



import socketserver
from dataclasses import dataclass

from textual.message import Message


@dataclass
class ShowURL(Message):
    url: str


@dataclass
class WebServerStarted(Message):
    server: socketserver.TCPServer


@dataclass
class WebServerStopped(Message):
    pass



from __future__ import annotations

from rich.text import Text
from textual import work
from textual.binding import Binding
from textual.events import MouseDown, MouseEvent, MouseMove, MouseUp
from textual.geometry import Offset
from textual.widget import Widget
from textual.widgets import Static

from par_infini_sweeper.data_structures import GameState, GridPos, SubGrid
from par_infini_sweeper.dialogs.highscore_dialog import HighscoreDialog
from par_infini_sweeper.dialogs.information import InformationDialog


class MainGrid(Widget, can_focus=True):
    """Widget that renders the infinite minesweeper grid and handles mouse interactions."""

    BINDINGS = [
        Binding(key="o", action="origin", description="Origin"),
        Binding(key="c", action="center", description="Center", show=False),
        Binding(key="d", action="debug", description="Debug", show=False),
        Binding(key="p", action="pause", description="Pause"),
        Binding(key="s", action="subgrid_highlight", description="Subgrid Highlight"),
        Binding(key="ctrl+d", action="xray", description="X-Ray", show=False),
    ]
    ALLOW_SELECT = False

    def __init__(self, game_state: GameState, info_bar: Static, debug_panel: Static) -> None:
        super().__init__()
        game_state.parent = self
        self.info_bar = info_bar
        self.debug_panel = debug_panel
        self.game_state: GameState = game_state
        self.drag_start: GridPos | None = None
        self.drag_threshold: int = 2  # minimal cells to distinguish a drag from a click
        self.is_dragging: bool = False
        self.debug = False
        self.debug_panel.display = self.debug
        self.mouse_sg: SubGrid | None = None

    def on_mount(self) -> None:
        if self.game_state.offset.is_origin:
            self.call_after_refresh(self.action_center)
        self.update_info()
        self.set_interval(1, self.update_info)

    def update_info(self) -> None:
        """Update the info bar with the current game state."""
        if not self.game_state.game_over and not self.game_state.paused:
            self.game_state.duration += 1

        game_over_text = "" if not self.game_state.game_over else " - [red]Game Over![/]"
        self.info_bar.update(
            " - ".join(
                [
                    f" [#5F5FFF]{self.game_state.user['nickname']}[/]",
                    f"Difficulty: [#00FF00]{self.game_state.difficulty}[/]",
                    f"Solved: {self.game_state.num_solved} / {self.game_state.num_subgrids}",
                    f"Score: [#00FF00]{self.game_state.score()}[/]",
                    f"Time: {self.game_state.time_played}{game_over_text}",
                ]
            )
        )
        self.debug_panel.update(
            "\n".join(
                [
                    f"NumHighlighted: {len(self.game_state.highlighted_cells)}",
                    f"NumSaved: {self.game_state.num_grids_saved}",
                    f"BoardOffset: {self.game_state.offset}",
                    f"BoardCenter: {self.game_state.compute_board_center()}",
                ]
            )
        )

    def render(self) -> Text:
        """Render the visible portion of the grid as text. Each cell is represented by two characters."""
        cells_x: int = self.size.width // 2  # each cell is 2 characters wide
        cells_y: int = self.size.height
        lines = [
            "".join(
                self.game_state.get_cell_representation(col + self.game_state.offset.x, row + self.game_state.offset.y)
                for col in range(cells_x)
            )
            for row in range(cells_y)
        ]
        return Text.from_markup("\n".join(lines))

    def action_center(self) -> None:
        """Center view on center of board"""
        c = self.game_state.compute_board_center()
        self.game_state.offset = Offset(-self.size.width // 4 + c.x, -self.size.height // 4 + c.y)
        self.game_state.save()
        self.refresh()

    def action_origin(self) -> None:
        """Center view on center of first subgrid"""
        self.game_state.offset = Offset(-self.size.width // 4 + 5, -self.size.height // 4 - 3)
        self.game_state.save()
        self.refresh()

    def action_debug(self) -> None:
        """Toggle the debug mode for the game."""
        self.debug = not self.debug
        self.debug_panel.display = self.debug

    def action_xray(self) -> None:
        """Toggle the x-ray mode for the game."""
        self.game_state.xray = not self.game_state.xray
        self.refresh()

    @work
    async def action_pause(self) -> None:
        self.game_state.paused = True
        await self.app.push_screen_wait(InformationDialog("Paused", "Press ESC to continue"))
        self.game_state.paused = False

    def handle_click(self, event: MouseDown | MouseUp) -> None:
        """
        Handle a click event by converting the event position to a global cell coordinate.
        Left-click reveals a cell; middle-click toggles a mark.

        Args:
            event (MouseDown | MouseUp): The mouse event
        """
        if self.is_dragging:
            return
        gx, gy = self.game_state.mouse_to_global_grid_coords(event)
        if event.button == 1 and not (event.shift or event.ctrl):
            self.game_state.reveal_cell(gx, gy)
            self.game_state.first_click = False
        elif event.button == 1 and (event.shift or event.ctrl):
            self.game_state.toggle_mark(gx, gy, True)

    def adjust_mouse_pos(self, event: MouseEvent) -> None:
        """
        Adjust the mouse position to account for padding and cell size.

        Args:
            event (MouseEvent): The mouse event

        """
        event._x -= self.styles.padding.width + 1
        event._y -= self.styles.padding.height + 1

    def on_mouse_down(self, event: MouseDown) -> None:
        """
        Record mouse down for drag detection.

        Args:
            event (MouseDown): The mouse down event
        """
        self.adjust_mouse_pos(event)
        self.drag_start = event.x, event.y

        if event.button == 1 and (event.shift or event.ctrl):
            gx, gy = self.game_state.mouse_to_global_grid_coords(event)
            self.game_state.highlight_neighbors(gx, gy)

    def on_mouse_move(self, event: MouseMove) -> None:
        """
        Handle mouse move events to detect dragging.

        Args:
            event (MouseMove): The mouse move event
        """
        self.adjust_mouse_pos(event)
        self.game_state.update_mouse_info(event)
        # cell = self.game_state.global_to_cell(self.sg_coord)
        # if cell:
        #     if self.mouse_sg != cell.parent:
        #         if self.mouse_sg:
        #             self.mouse_sg.highlighted = False
        #         self.mouse_sg = cell.parent
        #         self.mouse_sg.highlighted = True
        if self.drag_start is not None:
            dx: int = event.x - self.drag_start[0]
            dy: int = event.y - self.drag_start[1]
            if abs(dx) >= self.drag_threshold or abs(dy) >= self.drag_threshold:
                # Update the view offset in the opposite direction of the drag.
                self.game_state.offset -= Offset(dx, dy)
                self.drag_start = (event.x, event.y)
                self.is_dragging = True
                self.refresh()

    def on_mouse_up(self, event: MouseUp) -> None:
        """
        Check if mouse up is a click or drag and handle accordingly.

        Args:
            event (MouseUp): The mouse
        """
        # If the drag was minimal, treat this as a click.
        if self.game_state.game_over:
            self.drag_start = None
            self.is_dragging = False
            return

        self.adjust_mouse_pos(event)
        self.game_state.offset = self.game_state.offset
        if self.drag_start is not None:
            self.handle_click(event)
        self.drag_start = None
        self.is_dragging = False
        self.game_state.clear_highlighted()
        self.game_state.save()
        self.refresh()
        if self.game_state.game_over:
            self.app.push_screen(HighscoreDialog(self.game_state))

            self.app.push_screen(
                InformationDialog("Game Over", f"[red]You hit a mine.[/]\nScore: [yellow]{self.game_state.score()}")
            )

    def action_subgrid_highlight(self) -> None:
        self.game_state.highlighted_subgrid = not self.game_state.highlighted_subgrid
        if not self.game_state.highlighted_subgrid:
            self.refresh()



"""Provides an information dialog."""

from __future__ import annotations

from .text_dialog import TextDialog


class InformationDialog(TextDialog):
    """Modal dialog that shows information."""

    DEFAULT_CSS = """
    InformationDialog > Vertical {
        border: thick $primary 50%;
    }
    """



"""Data structures for Par Infinite Minesweeper game."""

from __future__ import annotations

import os
import random
import time
from typing import Any

import orjson
from authlib.integrations.requests_client import OAuth2Session
from jose import jwt
from textual.events import MouseEvent
from textual.geometry import Offset
from textual.widget import Widget

from par_infini_sweeper import db
from par_infini_sweeper.auth import build_auth_client
from par_infini_sweeper.db import get_db_connection, get_user
from par_infini_sweeper.enums import GameDifficulty, GameMode
from par_infini_sweeper.models import ChangeNicknameRequest, ChangeNicknameResponse, PostScoreRequest, PostScoreResult
from par_infini_sweeper.utils import format_duration

GridPos = tuple[int, int]

mine_counts: dict[GameDifficulty, int] = {GameDifficulty.EASY: 8, GameDifficulty.MEDIUM: 12, GameDifficulty.HARD: 16}
difficulty_mult: dict[GameDifficulty, int] = {GameDifficulty.EASY: 1, GameDifficulty.MEDIUM: 2, GameDifficulty.HARD: 3}

# Color mapping based on the count of adjacent mines
count_to_color: dict[int, str] = {
    0: "#FFFFFF",
    1: "#00FF00",
    2: "#5050FF",
    3: "#9327FF",
    4: "#FF00FF",
    5: "#FFFF00",
    6: "#FFA600",
    7: "#FF0000",
    8: "#A00000",
}


class Cell:
    """Represents a single cell in a subgrid."""

    def __init__(self, parent: SubGrid, is_mine: bool, marked: bool = False, uncovered: bool = False) -> None:
        self._parent = parent
        self._is_mine: bool = is_mine
        self._marked: bool = marked
        self._uncovered: bool = uncovered
        self._changed: bool = False
        self._highlighted: bool = False

    @property
    def is_mine(self) -> bool:
        return self._is_mine

    @is_mine.setter
    def is_mine(self, value: bool) -> None:
        if self._is_mine != value:
            self._is_mine = value
            self.changed = True

    @property
    def marked(self) -> bool:
        return self._marked

    @marked.setter
    def marked(self, value: bool) -> None:
        if self._marked != value:
            self._marked = value
            self.changed = True

    @property
    def uncovered(self) -> bool:
        return self._uncovered

    @uncovered.setter
    def uncovered(self, value: bool) -> None:
        if self._uncovered != value:
            self._uncovered = value
            self.changed = True

    @property
    def highlighted(self) -> bool:
        return self._highlighted

    @highlighted.setter
    def highlighted(self, value: bool) -> None:
        if self._highlighted != value:
            self._highlighted = value
            self.parent.parent.highlighted_cells.add(self) if value else self.parent.parent.highlighted_cells.discard(
                self
            )

    @property
    def parent(self) -> SubGrid:
        return self._parent

    @property
    def changed(self) -> bool:
        return self._changed

    @changed.setter
    def changed(self, value: bool) -> None:
        if self._changed != value:
            self._changed = value
            if value:
                self._parent.changed = True

    def to_dict(self) -> dict[str, bool]:
        """Return a dictionary representation of this cell."""
        return {"is_mine": self.is_mine, "marked": self.marked, "uncovered": self.uncovered}

    @staticmethod
    def from_dict(parent: SubGrid, data: dict[str, bool]) -> Cell:
        """Create a Cell instance from its dictionary representation."""
        return Cell(parent, data["is_mine"], data["marked"], data["uncovered"])


class SubGrid:
    """Represents an 8×8 subgrid of cells."""

    def __init__(
        self,
        parent: GameState,
        pos: GridPos,
        difficulty: GameDifficulty | None = None,
    ) -> None:
        """
        Initialize a subgrid at position `pos` with the given difficulty.
        Optionally ensure that the cell at local coordinate safe_pos is mine‑free.

        Args:
            parent (GameState): The parent game state.
            pos (GridPos): The position of the subgrid.
            difficulty (GameDifficulty | None): The difficulty level of the game.
        """
        self._changed: bool = False
        self.pos: GridPos = pos
        self.cells: list[list[Cell]] = self.generate_cells(difficulty) if difficulty else []
        self.solved: bool = False
        self._parent: GameState = parent

    @property
    def parent(self) -> GameState:
        return self._parent

    @property
    def changed(self) -> bool:
        return self._changed

    @changed.setter
    def changed(self, value: bool) -> None:
        if self._changed != value:
            self._changed = value
            if value:
                self.parent.changed_subgrids.add(self)

    def generate_cells(self, difficulty: GameDifficulty) -> list[list[Cell]]:
        """
        Generate an 8×8 grid of cells with mines distributed according to difficulty.
        The number of mines per subgrid is adjusted as follows:
          - easy: 8 mines
          - medium: 12 mines
          - hard: 16 mines
        """

        num_mines: int = mine_counts.get(difficulty, 8)
        positions: list[GridPos] = [(x, y) for y in range(8) for x in range(8)]
        mine_positions: list[GridPos] = random.sample(positions, num_mines)
        grid: list[list[Cell]] = []
        for y in range(8):
            row: list[Cell] = []
            for x in range(8):
                is_mine: bool = (x, y) in mine_positions
                row.append(Cell(self, is_mine))
            grid.append(row)
        return grid

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of this subgrid."""
        return {
            "pos": self.pos,
            "cells": [[cell.to_dict() for cell in row] for row in self.cells],
            "solved": self.solved,
        }

    def clear_changed(self) -> None:
        """Clear the changed flag for all cells in the subgrid."""
        for row in self.cells:
            for cell in row:
                cell.changed = False
        self.changed = False

    @property
    def key_str(self) -> str:
        """Return a unique key for this subgrid."""
        return f"{self.pos[0]},{self.pos[1]}"

    @staticmethod
    def from_dict(parent: GameState, data: dict[str, Any]) -> SubGrid:
        """Create a SubGrid instance from its dictionary representation."""
        sg: SubGrid = SubGrid(parent, tuple(data["pos"]))
        sg.cells = [[Cell.from_dict(sg, cell) for cell in row] for row in data["cells"]]
        sg.solved = data.get("solved", False)
        if sg.solved:
            return sg
        # Ensure that the subgrid is solved if all non-mine cells are uncovered.
        for row in sg.cells:
            for cell in row:
                if not cell.is_mine and not cell.uncovered:
                    return sg
        sg.solved = True

        for row in sg.cells:
            for cell in row:
                if cell.is_mine and not cell.marked:
                    cell.marked = True

        return sg


class GameState:
    """Represents the overall game state including difficulty and all subgrids."""

    def __init__(self, parent: Widget | None, user: dict[str, Any]) -> None:
        self.parent = parent
        self.mode: GameMode = GameMode.INFINITE
        self.user: dict[str, Any] = user
        self.difficulty: GameDifficulty = user["prefs"]["difficulty"]
        self.theme: str = user["prefs"]["theme"]
        self.subgrids: dict[GridPos, SubGrid] = {}
        self.subgrids[(0, 0)] = SubGrid(self, (0, 0), self.difficulty)
        game: dict[str, Any] = user["game"]
        offset: list[str] = game["board_offset"].split(",")
        assert len(offset) == 2
        self.offset = Offset(int(offset[0]), int(offset[1]))
        self.num_solved: int = 0
        self.started_ts: int = int(time.time())
        self.duration: int = game["duration"]
        self.game_over: bool = game["game_over"]
        self.num_grids_saved: int = 0
        self.highlighted_cells: set[Cell] = set()
        self.changed_subgrids: set[SubGrid] = set()
        self.mouse_grid: SubGrid | None = None
        self.paused: bool = False
        self.xray: bool = False
        self._auth_client: OAuth2Session | None = None
        self.first_click: bool = True
        self.mouse_pos: GridPos = 0, 0
        self.mouse_global_grid_coord: GridPos = 0, 0
        self.mouse_sg_coord: GridPos = (self.mouse_global_grid_coord[0] // 8, self.mouse_global_grid_coord[1] // 8)
        self.shift_pressed: bool = False
        self.ctrl_pressed: bool = False
        self.highlighted_subgrid: bool = False

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of the game state."""
        return {
            "user_id": self.user["id"],
            "theme": self.theme,
            "difficulty": self.difficulty,
            "game_over": self.game_over,
            "started_ts": self.started_ts,
            "duration": self.duration,
            "offset": (self.offset.x, self.offset.y),
            "subgrids": {f"{k[0]},{k[1]}": sg.to_dict() for k, sg in self.subgrids.items()},
            "first_click": self.first_click,
        }

    def mouse_to_global_grid_coords(self, event: MouseEvent) -> GridPos:
        """
        Convert mouse event coordinates to global cell coordinates.

        Args:
            event (MouseEvent): The mouse event

        Returns:
            GridPos: The global cell coordinates (gx, gy)
        """

        gx: int = (event.x // 2) + self.offset.x
        gy: int = event.y + self.offset.y
        return gx, gy

    def update_mouse_info(self, event: MouseEvent):
        self.mouse_pos = event.x, event.y
        self.mouse_global_grid_coord = self.mouse_to_global_grid_coords(event)
        self.mouse_sg_coord = (self.mouse_global_grid_coord[0] // 8, self.mouse_global_grid_coord[1] // 8)

        if self.parent and self.highlighted_subgrid:
            self.parent.refresh()

        self.shift_pressed = event.shift
        self.ctrl_pressed = event.ctrl

    @property
    def auth_client(self) -> OAuth2Session:
        if not self._auth_client or not self.is_logged_in():
            assert self.parent and self.parent.app
            self._auth_client = build_auth_client(self.user, self.parent.app)
        return self._auth_client

    def is_logged_in(self, grace_time: int = 60) -> bool:
        """
        Check if the user is logged in by checking the access token.
        Args:
            grace_time (int): Grace period in seconds for token expiration.
        Returns:
            bool: True if the user is logged in and the access token is valid, False otherwise.
        """
        if not self.user["access_token"]:
            return False
        try:
            unverified_claims = jwt.get_unverified_claims(self.user["access_token"])
            time_remaining = unverified_claims.get("exp", 0) - int(time.time())
            # make sure we have a least grace_time seconds left on the token
            is_expired = time_remaining > grace_time
            return not is_expired or len(self.user["refresh_token"]) > 0
        except Exception as _:
            self.logout()
            return False

    def logout(self) -> None:
        """
        Clear the user's tokens from the database.
        """
        self.user["id_token"] = ""
        self.user["access_token"] = ""
        self.user["refresh_token"] = ""
        self.user["expires_at"] = 0
        with get_db_connection() as conn:
            cursor = conn.cursor()
            cursor.execute(
                """UPDATE users SET id_token = '', access_token = '', refresh_token = '', expires_at = 0 WHERE id = ?""",
                (self.user["id"],),
            )
        self._auth_client = None

    def save_user(self) -> None:
        """
        Save user data to the database.
        """
        db.save_user(self.user)

    def change_internet_nickname(self, nickname: str) -> ChangeNicknameResponse:
        """
        Change nickname on the internet leaderboard.
        Args:
            nickname (str): New nickname to set.

        Returns:
            PostScoreResult: Result of the post operation.
        """
        if not self.is_logged_in():
            raise Exception("User is not logged in")
        url = os.environ.get("PIM_LEADERBOARD_URL", "https://pim.pardev.net") + "/nickname"

        res = self.auth_client.post(url, data=ChangeNicknameRequest(nickname=nickname).model_dump_json()).json()
        if "detail" in res:
            raise Exception(res["detail"] or "Unknown server error")

        result = ChangeNicknameResponse.model_validate(res)
        if not result.status == "success":
            raise Exception(result.message or "Unknown server error")
        self.user["net_nickname"] = nickname
        self.save_user()
        return result

    def new_game(self) -> None:
        """Start a new game by resetting the game state."""

        sg = SubGrid(self, (0, 0), self.difficulty)
        sg.changed = True
        self.subgrids = {(0, 0): sg}
        self.offset = Offset(0, 0)
        self.num_solved = 0
        self.started_ts = int(time.time())
        self.duration = 0
        self.num_grids_saved = 0
        self.game_over = False
        self.clear_highlighted()
        self.clear_changed()
        self.xray = False
        self.first_click = True

        conn = db.get_db_connection()
        with conn:
            cursor = conn.cursor()
            # Update user preferences.
            cursor.execute(
                """DELETE FROM grids WHERE user_id = ?""",
                (self.user["id"],),
            )
        self.save()

    def compute_board_center(self) -> Offset:
        """Compute the center of the game board in cells based on subgrid positions."""
        c = (0, 0)
        for k in self.subgrids.keys():
            c = (c[0] + k[0] * 8, c[1] + k[1] * 8)

        return Offset(c[0] // len(self.subgrids) + 5, c[1] // len(self.subgrids) - 3)

    @property
    def num_changed(self) -> int:
        return sum(1 for sg in self.subgrids.values() if sg.changed)

    @property
    def num_subgrids(self) -> int:
        return len(self.subgrids)

    @staticmethod
    def load(parent: Widget | None, user_name: str, nickname: str | None = None) -> GameState:
        """
        Load the game state from the SQLite database or create a new one.

        Args:
            parent (Widget): The parent widget.
            user_name (str): The name of the user.
            nickname (str | None): The nickname of the user.
        """
        with db.get_db_connection() as conn:
            user = get_user(conn, user_name, nickname)
            user_id = user["id"]
            game = user["game"]
            state = GameState(parent, user)

            cursor = conn.cursor()
            cursor.execute(
                "SELECT * FROM grids WHERE game_id = ? AND user_id = ?",
                (
                    game["id"],
                    user_id,
                ),
            )

            state.num_solved = 0
            row = cursor.fetchone()
            while row:
                key_parts: list[str] = row["sub_grid_id"].split(",")
                coords: GridPos = (int(key_parts[0]), int(key_parts[1]))
                sg_data = orjson.loads(row["grid_data"])
                sg: SubGrid = SubGrid.from_dict(state, sg_data)
                state.subgrids[coords] = sg
                if sg.solved:
                    state.num_solved += 1
                for row in sg.cells:
                    if not state.first_click:
                        break
                    for cell in row:
                        if cell.uncovered:
                            state.first_click = False
                            break
                row = cursor.fetchone()

        return state

    def score(self) -> int:
        """Calculate the score based on the number of solved subgrids and difficulty."""
        return self.num_solved * mine_counts.get(self.difficulty, 8)

    def save_score(self) -> None:
        score = self.score()
        if score == 0:
            return
        conn = db.get_db_connection()
        with conn:
            cursor = conn.cursor()
            cursor.execute(
                """INSERT INTO highscores (game_id, user_id, score) VALUES (?, ?,?)""",
                (self.user["game"]["id"], self.user["id"], score),
            )

    def save(self) -> int:
        """
        Save the game state to the SQLite database.

        Returns:
            int: The number of subgrids saved.
        """
        with db.get_db_connection() as conn:
            self.user["prefs"] = {"theme": self.theme, "difficulty": self.difficulty}
            user_id = self.user["id"]
            self.user["game"]["duration"] = self.duration
            self.user["game"]["game_over"] = self.game_over
            self.user["game"]["board_offset"] = f"{self.offset.x},{self.offset.y}"

            cursor = conn.cursor()
            cursor.execute(
                """UPDATE user_prefs SET theme = ?, difficulty = ? WHERE id = ?""",
                (self.theme, self.difficulty.value, user_id),
            )
            cursor.execute(
                """UPDATE games SET game_over = ?, board_offset = ?, duration = ? WHERE user_id = ?""",
                (
                    self.user["game"]["game_over"],
                    self.user["game"]["board_offset"],
                    self.user["game"]["duration"],
                    user_id,
                ),
            )

            # Save each subgrid using upsert.
            self.num_grids_saved = 0
            for sg in self.changed_subgrids:
                self.num_grids_saved += 1
                sg.clear_changed()
                sub_grid_id = f"{sg.pos[0]},{sg.pos[1]}"
                grid_data = orjson.dumps(sg.to_dict()).decode("utf-8")
                cursor.execute(
                    """INSERT OR REPLACE INTO grids (game_id, user_id, sub_grid_id, grid_data) VALUES (?,?,?,?)""",
                    (self.user["game"]["id"], user_id, sub_grid_id, grid_data),
                )
            self.clear_changed()
        return self.num_grids_saved

    @property
    def time_played(self) -> str:
        """Calculate the time played in a human-readable format."""
        return format_duration(self.duration)

    def clear_highlighted(self) -> None:
        """Clear the highlighted flag for all cells in all subgrids."""
        for cell in list(self.highlighted_cells):
            cell.highlighted = False
        self.highlighted_cells.clear()

    def clear_changed(self) -> None:
        """Clear the changed flag for all subgrids."""
        for sg in list(self.changed_subgrids):
            sg.clear_changed()
        self.changed_subgrids.clear()

    def global_to_cell(self, gx: int, gy: int, create_if_needed: bool = False) -> Cell | None:
        """
        Convert global coordinates to local cell coordinates.

        Args:
            gx (int): The global x-coordinate of the cell
            gy (int): The global y-coordinate of the cell
            create_if_needed (bool): Whether to create the subgrid if it does not exist

        Returns:
            Cell | None: The cell at the given coordinates, or None if the subgrid does not exist and create_if_needed is False
        """
        sg_coord: GridPos = (gx // 8, gy // 8)
        local_x: int = gx % 8
        local_y: int = gy % 8
        if sg_coord not in self.subgrids:
            if not create_if_needed:
                return None
            self.subgrids[sg_coord] = SubGrid(self, sg_coord, self.difficulty)
            self.subgrids[sg_coord].changed = True
        subgrid: SubGrid = self.subgrids[sg_coord]
        return subgrid.cells[local_y][local_x]

    def cell_has_uncovered_neighbor(self, gx: int, gy: int) -> bool:
        """
        Check whether the cell at (gx, gy) has at least one uncovered neighbor.

        Args:
            gx (int): The global x-coordinate of the cell
            gy (int): The global y-coordinate of the cell

        Returns:
            bool: True if the cell has at least one uncovered neighbor, False otherwise
        """
        for dx in [-1, 0, 1]:
            for dy in [-1, 0, 1]:
                if dx == 0 and dy == 0:
                    continue
                nx: int = gx + dx
                ny: int = gy + dy
                n_sg: GridPos = (nx // 8, ny // 8)
                if n_sg in self.subgrids:
                    subgrid: SubGrid = self.subgrids[n_sg]
                    lx: int = nx % 8
                    ly: int = ny % 8
                    if subgrid.cells[ly][lx].uncovered or subgrid.cells[ly][lx].is_mine:
                        return True
        return False

    def count_adjacent_flags_mines(self, gx: int, gy: int) -> tuple[int, int]:
        """
        Count the number of mines adjacent to the cell at (gx, gy). Generates any adjacent subgrid that has not yet been created.

        Args:
            gx (int): The global x-coordinate of the cell
            gy (int): The global y-coordinate of the cell

        Returns:
            tuple[int, int]: A tuple containing the number of flagged cells and the number of mines

        """
        mine_count: int = 0
        flag_count: int = 0
        for dx in [-1, 0, 1]:
            for dy in [-1, 0, 1]:
                if dx == 0 and dy == 0:
                    continue
                nx: int = gx + dx
                ny: int = gy + dy
                n_sg: GridPos = (nx // 8, ny // 8)
                if n_sg not in self.subgrids:
                    continue
                subgrid: SubGrid = self.subgrids[n_sg]
                lx: int = nx % 8
                ly: int = ny % 8
                if subgrid.cells[ly][lx].is_mine:
                    mine_count += 1
                if subgrid.cells[ly][lx].marked:
                    flag_count += 1
        return flag_count, mine_count

    def reveal_cell(self, gx: int, gy: int, depth: int = 0) -> None:
        """
        Reveal the cell at global coordinates (gx, gy). If it is a mine the game ends. Also, generate any adjacent subgrids as needed.

        Args:
            gx (int): The global x-coordinate of the cell
            gy (int): The global y-coordinate of the cell
            depth (int): The depth of the recursion (to prevent infinite recursion)
        """
        if self.game_over:
            return
        sg_coord: GridPos = (gx // 8, gy // 8)
        assert self.parent
        # For non-initial subgrids, only allow a reveal if at least one neighbor is uncovered.
        if sg_coord != (0, 0) and not self.cell_has_uncovered_neighbor(gx, gy):
            self.parent.notify("No uncovered neighbors")
            return

        cell: Cell | None = self.global_to_cell(gx, gy, True)
        # Do nothing if cell is marked or uncovered.
        if not cell or cell.marked or cell.uncovered:
            return

        cell.uncovered = True
        if cell.uncovered and cell.is_mine:
            if not self.first_click:
                self.game_over = True
                self.save()
                self.save_score()
                self.parent.refresh()
                return
            cell.is_mine = False
            # move mine to a surrounding cell
            border_cell: Cell | None = None
            for dx in [-1, 0, 1]:
                if border_cell:
                    break
                for dy in [-1, 0, 1]:
                    if dx == 0 and dy == 0:
                        continue
                    nx: int = gx + dx
                    ny: int = gy + dy
                    border_cell: Cell | None = self.global_to_cell(nx, ny, True)
                    if border_cell and not border_cell.is_mine:
                        border_cell.is_mine = True
                        break
                    border_cell = None
        # Generate any adjacent subgrids.
        for dx in [-1, 0, 1]:
            for dy in [-1, 0, 1]:
                if dx == 0 and dy == 0:
                    continue
                nx: int = gx + dx
                ny: int = gy + dy
                n_sg: GridPos = (nx // 8, ny // 8)
                if n_sg not in self.subgrids:
                    self.subgrids[n_sg] = SubGrid(self, n_sg, self.difficulty)
                    self.subgrids[n_sg].changed = True

        # Prevent infinite recursion by limiting depth.
        if depth < 250:
            # If the cell has 0 neighboring mines, recursively reveal its neighbors.
            counts: tuple[int, int] = self.count_adjacent_flags_mines(gx, gy)
            if counts[1] == 0:
                for dx in [-1, 0, 1]:
                    for dy in [-1, 0, 1]:
                        if dx == 0 and dy == 0:
                            continue
                        nx: int = gx + dx
                        ny: int = gy + dy
                        self.reveal_cell(nx, ny, depth + 1)
        self.check_subgrid_solved(sg_coord)

    def reveal_surround(self, gx: int, gy: int) -> None:
        """
        Reveal surrounding cells if the cell at (gx, gy) is uncovered and the number of flagged cells matches the number of mines.

        Args:
            gx (int): The global x-coordinate of the cell
            gy (int): The global y-coordinate of the cell

        """
        if self.game_over:
            return
        cell: Cell | None = self.global_to_cell(gx, gy)
        if not cell or not cell.uncovered:
            return
        counts: tuple[int, int] = self.count_adjacent_flags_mines(gx, gy)
        if not counts[0]:
            return

        if cell.uncovered:
            if counts[0] != counts[1]:
                # self.notify("Flag count does not match mine count", severity="error")
                return

            for dx in [-1, 0, 1]:
                for dy in [-1, 0, 1]:
                    if dx == 0 and dy == 0:
                        continue
                    nx: int = gx + dx
                    ny: int = gy + dy
                    self.reveal_cell(nx, ny)
        self.save()
        assert self.parent

        self.parent.refresh()

    def highlight_neighbors(self, gx: int, gy: int) -> None:
        """
        Highlight the cell neighbors around (gx, gy).
        Args:
            gx (int): The global x-coordinate of the cell
            gy (int): The global y-coordinate of the cell
        """
        if self.game_over:
            return
        cell: Cell | None = self.global_to_cell(gx, gy)
        if not cell or not cell.uncovered:
            return
        counts: tuple[int, int] = self.count_adjacent_flags_mines(gx, gy)
        # if no neighboring mines do not highlight
        if not counts[1]:
            return

        for dx in [-1, 0, 1]:
            for dy in [-1, 0, 1]:
                if dx == 0 and dy == 0:
                    continue
                cell = self.global_to_cell(gx + dx, gy + dy)
                if cell and not cell.uncovered:
                    cell.highlighted = True
        assert self.parent
        self.parent.refresh()

    def toggle_mark(self, gx: int, gy: int, surround: bool = False) -> None:
        """
        Toggle the mark (flag) on the cell at global coordinates (gx, gy).

        Args:
            gx (int): The global x-coordinate of the cell
            gy (int): The global y-coordinate of the cell
            surround (bool): Whether to reveal surrounding cells if the cell is already uncovered
        """
        cell: Cell | None = self.global_to_cell(gx, gy)
        if not cell:
            return
        if cell.uncovered:
            if not surround:
                return
            self.reveal_surround(gx, gy)
            return
        cell.marked = not cell.marked
        self.save()

        assert self.parent
        self.parent.refresh()

    def check_subgrid_solved(self, sg_coord: GridPos) -> None:
        """
        Mark a subgrid as solved if the all cells that dont contain a mine have been uncovered.

        Args:
            sg_coord (GridPos): The coordinates of the subgrid
        """
        subgrid: SubGrid = self.subgrids[sg_coord]
        if subgrid.solved:
            return
        for row in subgrid.cells:
            for cell in row:
                if not cell.is_mine and not cell.uncovered:
                    return
        subgrid.solved = True
        self.num_solved += 1
        for row in subgrid.cells:
            for cell in row:
                if cell.is_mine and not cell.marked:
                    cell.marked = True

    def get_cell_representation(self, gx: int, gy: int) -> str:
        """
        Return a string representation of the cell at global coordinates (gx, gy).

        Args:
            gx (int): The global x-coordinate of the cell
            gy (int): The global y-coordinate of the cell

        Returns:
            str: The string representation of the cell
        """
        cell: Cell | None = self.global_to_cell(gx, gy)

        # set background based on checker pattern
        sg_coord: GridPos = (gx // 8, gy // 8)
        bg_color = "#000000" if (sg_coord[0] + sg_coord[1]) % 2 == 0 else "#111111"
        if self.mouse_sg_coord == sg_coord and self.highlighted_subgrid:
            bg_color = "#888800"

        if not cell:
            return f"[#C0C0C0 on {bg_color}]? [/]"  # placeholder for not-yet generated subgrid

        if self.xray or cell.uncovered:
            if cell.is_mine:
                if self.xray and cell.marked:
                    return f"[#FF0000 on {bg_color}]⚑ [/]"
                return f"[#FF0000 on {bg_color}]💣[/]"
            count: tuple[int, int] = self.count_adjacent_flags_mines(gx, gy)
            if cell.parent.solved:
                color = "#A0A0A0"
            else:
                color = "#FFFF00" if cell.highlighted else count_to_color.get(count[1], "#FFFFFF")
            if cell.parent.solved:
                if count[1] == 0:
                    return f"[{color} on {bg_color}]. [/]"
            if count[1] > 0:
                return f"[{color} on {bg_color}]{count[1]} [/]"
            return f"[#C0C0C0 on {bg_color}]  [/]"
        else:
            if cell.marked:
                return f"[#FF0000 on {bg_color}]⚑ [/]"
            color = "#FFFF00" if cell.highlighted else "#E0E0E0"
            return f"[{color} on {bg_color}]■ [/]"

    def post_internet_score(self) -> PostScoreResult:
        """
        Post score to the internet leaderboard.
        Args:
            game_state (GameState): Game state containing user and score data.
        Returns:
            PostScoreResult: Result of the post operation.
        """
        if not self.is_logged_in():
            raise Exception("User is not logged in")
        url = os.environ.get("PIM_LEADERBOARD_URL", "https://pim.pardev.net") + "/score"

        res = self.auth_client.post(
            url,
            data=PostScoreRequest(
                mode=self.mode.value,
                difficulty=self.difficulty.value,
                score=self.score(),
                duration=self.duration,
            ).model_dump_json(),
        ).json()

        if "detail" in res:
            raise Exception(res["detail"] or "Unknown server error")

        return PostScoreResult.model_validate(res)



import base64
import os
import sqlite3
from pathlib import Path
from sqlite3 import Connection, Cursor
from typing import Any

import requests
from xdg_base_dirs import xdg_data_home

from par_infini_sweeper import __application_binary__
from par_infini_sweeper.db_migrations import migrate_db_to_1_1, migrate_legacy_db
from par_infini_sweeper.enums import GameDifficulty, GameMode
from par_infini_sweeper.models import (
    ScoreData,
    ScoreDataResponse,
)

db_folder_old = Path(f"~/.{__application_binary__}").expanduser()
db_folder = xdg_data_home() / __application_binary__

if not db_folder.parent.exists():
    db_folder.parent.mkdir(parents=True, exist_ok=True)


if db_folder_old.exists() and not db_folder.exists():
    db_folder = db_folder_old.rename(db_folder)


db_path = db_folder / "game_data.sqlite"
db_bak_path = db_folder / "game_data.sqlite.bak"


def get_db_connection() -> sqlite3.Connection:
    """
    Return a connection to the SQLite database with a 5 sec timeout.
    Also creates a backup of the database if it doesn't exist or if it's older than 1 day.
    """
    db_folder.mkdir(parents=True, exist_ok=True)
    if db_path.exists():
        if not db_bak_path.exists():
            db_bak_path.write_bytes(db_path.read_bytes())
        else:
            # if backup is older than 1 day, replace it with the current db
            if db_bak_path.stat().st_mtime < (db_path.stat().st_mtime - 86400):
                db_bak_path.write_bytes(db_path.read_bytes())

    conn = sqlite3.connect(db_path, timeout=5)
    conn.row_factory = sqlite3.Row
    return conn


def init_db(conn: Connection, username: str = "user", nickname: str | None = None) -> None:
    """
    Initialize the SQLite database with required tables and default user.

    Args:
        conn (Connection): SQLite connection object.
        username (str): Default username.
        nickname (str | None): Default nickname.

    """

    if len(username) > 20:
        raise ValueError("Username must be 20 characters or less")

    if nickname and len(nickname) > 20:
        raise ValueError("Nickname must be 20 characters or less")

    conn = conn or get_db_connection()
    with conn:
        cursor = conn.cursor()

        # Check if pim_db_info and users tables exist
        cursor.execute("SELECT name FROM sqlite_master WHERE type='table' AND name='pim_db_info'")
        pim_db_info_exists = cursor.fetchone() is not None

        cursor.execute("SELECT name FROM sqlite_master WHERE type='table' AND name='users'")
        users_exists = cursor.fetchone() is not None

        cursor.execute("""
            CREATE TABLE IF NOT EXISTS pim_db_info (
                version TEXT PRIMARY KEY
            )
        """)

        # update usernames and nicknames to be no longer than 30 characters
        if users_exists:
            cursor.execute("UPDATE users SET username = substr(username, 1, 30), nickname = substr(nickname, 1, 30)")

        cursor.execute("""
            CREATE TABLE IF NOT EXISTS users (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                username TEXT UNIQUE NOT NULL,
                nickname TEXT UNIQUE NOT NULL,
                net_nickname TEXT NOT NULL DEFAULT '',
                id_token TEXT DEFAULT '',
                access_token TEXT DEFAULT '',
                refresh_token TEXT DEFAULT '',
                expires_at INTEGER NOT NULL DEFAULT 0,
                created_ts TIMESTAMP DEFAULT CURRENT_TIMESTAMP
            )
        """)
        cursor.execute("""
            CREATE TABLE IF NOT EXISTS user_prefs (
                id INTEGER PRIMARY KEY,
                theme TEXT NOT NULL,
                difficulty TEXT NOT NULL CHECK(difficulty IN ('easy','medium','hard')),
                FOREIGN KEY(id) REFERENCES users(id) ON DELETE CASCADE
            )
        """)
        cursor.execute("""
            CREATE TABLE IF NOT EXISTS games (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                user_id INTEGER NOT NULL,
                mode TEXT NOT NULL DEFAULT 'infinite',
                game_over BOOLEAN NOT NULL DEFAULT 0,
                duration INTEGER NOT NULL DEFAULT 0,
                board_offset TEXT NOT NULL DEFAULT '0,0',
                created_ts TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                FOREIGN KEY(user_id) REFERENCES users(id) ON DELETE CASCADE
            )
        """)
        cursor.execute("""
            CREATE TABLE IF NOT EXISTS highscores (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                user_id INTEGER NOT NULL,
                game_id INTEGER NOT NULL,
                score INTEGER NOT NULL,
                created_ts TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                FOREIGN KEY(user_id) REFERENCES users(id) ON DELETE CASCADE,
                FOREIGN KEY(game_id) REFERENCES games(id) ON DELETE CASCADE
            )
        """)
        cursor.execute("""
            CREATE TABLE IF NOT EXISTS grids (
                game_id INTEGER NOT NULL,
                user_id INTEGER NOT NULL,
                sub_grid_id TEXT NOT NULL,
                grid_data TEXT NOT NULL,
                FOREIGN KEY(user_id) REFERENCES users(id) ON DELETE CASCADE,
                FOREIGN KEY(game_id) REFERENCES games(id) ON DELETE CASCADE,
                PRIMARY KEY (game_id, user_id, sub_grid_id)
            )
        """)

        # If users table exists but pim_db_info does not, run migrate_db
        if users_exists and not pim_db_info_exists:
            migrate_legacy_db(conn)

        cursor.execute("SELECT version FROM pim_db_info")
        db_version = cursor.fetchone()
        if db_version:
            db_version = db_version[0]
        if db_version is None:
            db_version = "1.0"
            cursor.execute("INSERT INTO pim_db_info (version) VALUES (?)", (db_version,))
        if db_version == "1.0":
            migrate_db_to_1_1(conn)
            db_version = "1.1"

        # Create default user "user" if not exists.
        cursor.execute("SELECT id FROM users WHERE username = ?", (username,))
        if cursor.fetchone() is None:
            cursor.execute(
                "INSERT INTO users (username, nickname) VALUES (?, ?)", (username, nickname or username.capitalize())
            )
            user_id = cursor.lastrowid
            cursor.execute(
                "INSERT INTO user_prefs (id, theme, difficulty) VALUES (?,?,?)", (user_id, "textual-dark", "easy")
            )
            cursor.execute("INSERT INTO games (user_id) VALUES (?)", (user_id,))


def get_user(conn: Connection, username: str = "user", nickname: str | None = None) -> dict[str, Any]:
    """
    Load or create requested user.

    Args:
        conn (Connection): SQLite connection object.
        username (str): Username to load or create.
        nickname (str | None): Nickname to load or create.

    Returns:
        dict[str, Any]: User data including preferences and game state.

    """
    cursor: Cursor = conn.cursor()
    cursor.execute("SELECT * FROM users WHERE username = ?", (username,))
    user = dict(cursor.fetchone())
    if user is None:
        # If default user not found, initialize the DB.
        init_db(conn, username, nickname)
        cursor.execute("SELECT id FROM users WHERE username = ?", ("user",))
        user = dict(cursor.fetchone())
    else:
        user["nickname"] = nickname or user["nickname"]
        cursor.execute("UPDATE users SET nickname = ? WHERE id = ?", (user["nickname"], user["id"]))

    user_id = user["id"]
    cursor.execute("SELECT theme, difficulty FROM user_prefs WHERE id = ?", (user_id,))
    prefs = dict(cursor.fetchone())
    prefs["difficulty"] = GameDifficulty(prefs["difficulty"])
    user["prefs"] = prefs

    cursor.execute("SELECT * FROM games WHERE user_id = ?", (user_id,))
    game = dict(cursor.fetchone())
    user["game"] = game

    cursor.execute(
        "SELECT * FROM highscores WHERE game_id=? AND user_id = ? order by created_ts limit 10",
        (
            game["id"],
            user_id,
        ),
    )
    highscores = cursor.fetchall()
    user["highscores"] = [dict(s) for s in highscores]

    return user


def get_highscores(num_scores: int = 10) -> dict[GameMode, list[dict[str, Any]]]:
    """
    Return top num_scores highscores for each mode.

    Args:
        num_scores (int): Number of top scores to return for each game mode.

    Returns:
        dict[GameMode,list[dict[str, Any]]]: List of top 10 highscore records for each game mode

    """
    if num_scores < 1:
        raise ValueError("num_scores must be at least 1")
    with get_db_connection() as conn:
        cursor = conn.cursor()
        cursor.execute(
            """
        WITH RankedScores AS (
            SELECT
                score,
                h.created_ts,
                u.nickname,
                g.mode,
                g.duration,
                ROW_NUMBER() OVER (PARTITION BY g.mode ORDER BY score DESC, h.created_ts DESC) as rank
            FROM highscores h
            JOIN users u ON h.user_id = u.id
            JOIN games g ON g.id = h.game_id
        )
        SELECT
            score,
            created_ts,
            nickname,
            mode,
            duration
        FROM RankedScores
        WHERE rank <= ?
        ORDER BY mode, rank;
        """,
            (num_scores,),
        )
        highscores = cursor.fetchall()
        highscores = [dict(s) for s in highscores]
        result: dict[GameMode, list[dict[str, Any]]] = {}
        for gm in GameMode:
            result[gm] = []
        for h in highscores:
            if h["mode"] not in result:
                result[GameMode(h["mode"])] = []
            result[GameMode(h["mode"])].append(h)
        return result


def get_internet_highscores() -> dict[GameMode, list[ScoreData]]:
    ret: dict[GameMode, list[ScoreData]] = {}
    for mode in GameMode:
        ret[mode] = []

    url = os.environ.get("PIM_LEADERBOARD_URL", "https://pim.pardev.net") + "/score"
    # yes i know this is bad, but it helps keep random bots from hitting the api
    headers = {"api-key": base64.b64decode("V29vdFdvb3RXb290").decode(encoding="utf-8")}
    res = requests.request("GET", url, headers=headers, timeout=5).json()
    if "detail" in res:
        raise Exception(res["detail"] or "Unknown server error")
    response: ScoreDataResponse = ScoreDataResponse.model_validate(res)
    if response.status == "error":
        raise Exception(response.message or "Unknown server error")
    for item in response.scores:
        ret[GameMode(item.mode)].append(item)

    return ret


def save_user(user: dict[str, Any]) -> None:
    """
    Save user data to the database.
    Args:
        user (dict[str, Any]): User data containing id_token and access_token.
    """
    with get_db_connection() as conn:
        cursor = conn.cursor()
        cursor.execute(
            """UPDATE users
            SET id_token = ?, access_token = ?, refresh_token = ?, expires_at = ?, nickname=?, net_nickname=?
            WHERE id = ?""",
            (
                user["id_token"],
                user["access_token"],
                user["refresh_token"],
                user["expires_at"],
                user["nickname"],
                user["net_nickname"],
                user["id"],
            ),
        )



def format_duration(elapsed: int) -> str:
    """Format elapsed time in seconds to HH:MM:SS"""
    hours, remainder = divmod(elapsed, 3600)
    minutes, seconds = divmod(remainder, 60)
    return f"{str(hours).rjust(2, '0')}:{str(minutes).rjust(2, '0')}:{str(seconds).rjust(2, '0')}"


def escape_brackets(v: str) -> str:
    """Escape square brackets in a string"""
    return v.replace("[", "\\[").replace("]", "\\]")
